//***************************************************************************
// CMG_LLIO_EMU_DISPLAY_GDI.c
//***************************************************************************
// CMG - Christian Merkle Graphics
// Copyright (C) 2oo6 - 2oo8 Christian Merkle 
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// To contact Christian Merkle, see his website at:
// http://www.cmerkle.de
//***************************************************************************
// Author:          Christian Merkle
// Created:         27.10.2006
//***************************************************************************
// History:
//   10.11.2006     *stable*
//   10.11.2006     added an interthread event to check for successful 
//                  window creation and thread startup
//   10.11.2006     added a safe _Exit with wait states
//   10.11.2006     added redraw only on demand
//   10.11.2006     added _SetMode support
//   10.11.2006     added extern _PixelColor support
//   12.11.2006     prepare lcd bitmap on _Init (clear all pixels)
//   23.01.2007     added power management
//***************************************************************************
// ToDo:
//   [ ]    xxx
//***************************************************************************


// main CMG include (will include user configuration)
#include    "../../CMG_Main.h"



//---------------------------------------------------------------------------
// conditional compilation and configuration checks
//---------------------------------------------------------------------------
#ifndef CMG_LLIO_CONFIGURED
    #error CMG_LLIO not configured. Please customize the config file correctly (use the *.config template).
#elif defined( CMG_LLIO_EMU_DISPLAY_GDI )

    #if !defined( CMG_LLIO_EMU_WIDTH ) || !defined( CMG_LLIO_EMU_HEIGHT ) || !defined( CMG_LLIO_EMU_BITDEPTH )
        #error CMG_LLIO_EMU_* are not defined. Please customize the config file correctly using the *.config template.
    #endif 
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// inludes
//---------------------------------------------------------------------------
// CMG:
#include    "../CMG_LLIO.h"
#include    "CMG_LLIO_EMU_DISPLAY.h"
#include    "CMG_LLIO_EMU_DISPLAY_COMMON.h"
// other:
#include    <windows.h>
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// defines
//---------------------------------------------------------------------------
#define     DISPLAY_WINDOW_CLASSNAME        "CMG_LLIO_EMU_DISPLAY_GDI"
#define     DISPLAY_WINDOW_TITLE            "CMG - LCD Emulator"
#define     DISPLAY_WINDOW_WAITTIME_MS      5000    // wait 5 secs
#define     DISPLAY_WINDOW_TIMER_MS         25      // every 25 msec (max 40fps)

#define     MAX_NUMBER_OF_ERRORS            10      // maximum space for 10 errors

#define     MENU_MODE_PHOTO                 1
#define     MENU_MODE_EXACT                 2
#define     MENU_MODE_PHOTO_TEXT            "Photo Mode"
#define     MENU_MODE_EXACT_TEXT            "Exact Mode"
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// globals
//---------------------------------------------------------------------------
HWND            g_DISPLAY_hWnd = NULL;          // lcd window handle
HANDLE          g_DISPLAY_hThread = NULL;       // window thread handle
HMENU           g_DISPLAY_hMenu = NULL;         // popup menu
UINT            g_DISPLAY_uTimerID = 0;         // timer id

cmg_u8         *g_DISPLAY_pbyLCDBuffer = NULL;  // local buffer
cmg_Coord       g_DISPLAY_iWidth;               // emulated lcd width in pixels
cmg_Coord       g_DISPLAY_iHeight;              // emulated lcd height in pixels

cmg_u32        *g_DISPLAY_pdwLCDBackground = NULL; // lcd background bitmap
BITMAPINFO      g_DISPLAY_pLCDBackground_BI;    // lcd bitmap bitmap info
cmg_Coord       g_DISPLAY_iLCDBackground_Width; // lcd bitmap size in pixels
cmg_Coord       g_DISPLAY_iLCDBackground_Height;// lcd bitmap size in pixels

cmg_u32        *g_DISPLAY_pdwLCD = NULL;        // lcd bitmap
BITMAPINFO      g_DISPLAY_pLCD_BI;              // lcd bitmap bitmap info
cmg_Coord       g_DISPLAY_iLCD_Width;           // lcd bitmap size in pixels
cmg_Coord       g_DISPLAY_iLCD_Height;          // lcd bitmap size in pixels

cmg_u32         g_DISPLAY_aadwPixelColor[256][4];// graycolor 2x2 pixel values

volatile cmg_bool   g_DISPLAY_bLCD_On;          // lcd mode status
volatile cmg_bool   g_DISPLAY_bModeChanged;     // content changed since last timer
volatile cmg_bool   g_DISPLAY_bContentChanged;  // content changed since last timer
volatile cmg_bool   g_DISPLAY_bBacklightOn;     // backlight status
volatile cmg_bool   g_DISPLAY_bExactMode;       // exact mode status

volatile cmg_u32    g_DISPLAY_dwErrors;         // number of errors (should be 0)
cmg_char            *(g_DISPLAY_astrError[MAX_NUMBER_OF_ERRORS]);  // maximum of 10 errors (allocated strings)
const cmg_char      g_DISPLAY_cstrMaxErrors[] = "Maximum number of error messages reached, further will be ignored...";
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------------
// predefs:
_PRIVATE    LRESULT     CALLBACK    _DISPLAY_WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
_PRIVATE    DWORD       WINAPI      _DISPLAY_ThreadProc( void *pParameter );
_PRIVATE    void                    _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer( void );
//---------------------------------------------------------------------------





/*---------------------------------------------------------------------------
| _DISPLAY_WndProc
|-----------------------------------
| LCD window callback handler
+----------------------------------------------------------------------------
| @Params:          default win32 (see sdk)
| @Return Value:    default win32 (see sdk)
+--------------------------------------------------------------------------*/
_PRIVATE
LRESULT CALLBACK _DISPLAY_WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    HDC             hDc;                // decice context to draw in
    PAINTSTRUCT     ps;                 // paintstruct for WM_PAINT

    switch ( uMsg )
    {
        // draw the background
        case WM_ERASEBKGND:
            hDc = (HDC)wParam;
            StretchDIBits( hDc, 0, 0, g_DISPLAY_iLCDBackground_Width, g_DISPLAY_iLCDBackground_Height, 0, 0, g_DISPLAY_iLCDBackground_Width, g_DISPLAY_iLCDBackground_Height, g_DISPLAY_pdwLCDBackground, &g_DISPLAY_pLCDBackground_BI, DIB_RGB_COLORS, SRCCOPY );
            return 1;


        // draw the lcd content
        case WM_PAINT:
            // get dc
            hDc = BeginPaint (hWnd, &ps);
            
            // if there are no errors
            if ( !g_DISPLAY_dwErrors )
            {
                // draw the lcd content
                if ( g_DISPLAY_bLCD_On )
                    StretchDIBits( hDc, CMG_LLIO_DISPLAY_COMMON_LCDBG_START_X, CMG_LLIO_DISPLAY_COMMON_LCDBG_START_Y, g_DISPLAY_iLCD_Width, g_DISPLAY_iLCD_Height, 0, 0, g_DISPLAY_iLCD_Width, g_DISPLAY_iLCD_Height, g_DISPLAY_pdwLCD, &g_DISPLAY_pLCD_BI, DIB_RGB_COLORS, SRCCOPY );
            }
            else
            {
                cmg_mu32        uIdx;
                TEXTMETRIC      tm;
                // on error output errors
                GetTextMetrics( hDc, &tm );
                SetTextColor( hDc, RGB( 128, 0, 0 ) );
                for ( uIdx = 0; uIdx < g_DISPLAY_dwErrors; uIdx++ )
                    TextOut( hDc, 30, uIdx * tm.tmHeight + 30, g_DISPLAY_astrError[ uIdx ], strlen( g_DISPLAY_astrError[ uIdx ] ) );
                if ( g_DISPLAY_dwErrors == MAX_NUMBER_OF_ERRORS )
                    TextOut( hDc, 30, uIdx * tm.tmHeight + 30, g_DISPLAY_cstrMaxErrors, strlen( g_DISPLAY_cstrMaxErrors ) );
            }

            // release
            EndPaint( hWnd, &ps );
            return 0;


        // periodically timer (DISPLAY_WINDOW_TIMER_MS == 25msec / 40fps)
        case WM_TIMER:
            // mode or content changed?
            if ( g_DISPLAY_bModeChanged )
            {
                // redraw lcd content AND background
                InvalidateRect( hWnd, NULL, true );
                g_DISPLAY_bModeChanged = false;
                g_DISPLAY_bContentChanged = false;
            }
            else if ( g_DISPLAY_bContentChanged )
            {
                // redraw lcd content
                InvalidateRect( hWnd, NULL, false );
                g_DISPLAY_bContentChanged = false;
            }
            return 0;


        // make window moveable on every point
        case WM_NCHITTEST:
            // just return that user picked the caption bar
            return HTCAPTION;


        // if right butten was pressed
        case WM_NCRBUTTONUP:
        {
            UINT    uSelected;
            POINTS  Points;

            // get mouse pos
            Points = MAKEPOINTS( lParam );

            // show popup
            uSelected = TrackPopupMenu( g_DISPLAY_hMenu, TPM_RETURNCMD, Points.x, Points.y, 0, hWnd, NULL );
            
            // select result
            if ( uSelected == MENU_MODE_PHOTO )
                g_DISPLAY_bExactMode = false;
            else if ( uSelected == MENU_MODE_EXACT )
                g_DISPLAY_bExactMode = true;

            // load and apply pixel color mode and redraw from buffer
            _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer();
            return 0;
        }


        // dont allow close
        case WM_CLOSE:
            return 0;


        // destroy window
        case WM_DESTROY:
            // destroy and free
            if ( g_DISPLAY_uTimerID )
                KillTimer( hWnd, g_DISPLAY_uTimerID );
            // reset handle
            g_DISPLAY_hWnd = NULL;
            // quit
            PostQuitMessage(0);
            return 0;


        // let DefWindowProc do the default stuff
        default:
            return DefWindowProc( hWnd, uMsg, wParam, lParam );
    }
    return 0;
}



/*---------------------------------------------------------------------------
| _DISPLAY_ThreadProc
|-----------------------------------
| LCD window thread
+----------------------------------------------------------------------------
| @Params:          default win32 (see sdk)
| @Return Value:    default win32 (see sdk)
+--------------------------------------------------------------------------*/
_PRIVATE
DWORD WINAPI _DISPLAY_ThreadProc( void *pParameter )
{
    cmg_Result  iResult         = CMG_OK;
    HINSTANCE   hInstance       = GetModuleHandle( NULL );
    WNDCLASS    wndClass;
    MSG         msg;
    int         iWindowX, iWindowY;
    HANDLE      hWaitEvent      = *(HANDLE*)pParameter;
    UINT        uTimerID = 0;


    // param check
    CHECK( hWaitEvent == NULL, CMG_ERROR_OS );


    // . . . . . . . . . . . . . . . . . . . . . . .
    // create class
    // set window class and register it
    ZeroMemory( &wndClass, sizeof(wndClass) );
    wndClass.style          = 0;                        // default
    wndClass.lpfnWndProc    = _DISPLAY_WndProc;         // our window callback handler
    wndClass.cbClsExtra     = 0;                        // nothing
    wndClass.cbWndExtra     = 0;                        // nothing
    wndClass.hInstance      = hInstance;                // instance
    wndClass.hIcon          = NULL;                     // no icon
    wndClass.hbrBackground  = NULL;                     // no background (we do it ourselve)
    wndClass.lpszMenuName   = NULL;                     // no menu
    wndClass.hCursor        = LoadCursor( NULL, IDC_ARROW );    // default curser
    wndClass.lpszClassName  = DISPLAY_WINDOW_CLASSNAME; // our classname
    CHECK( !RegisterClass( &wndClass ), CMG_ERROR_OS ) 

    

    // . . . . . . . . . . . . . . . . . . . . . . .
    // create and show main window
    // calculate startup position (centered)
    iWindowX = ( GetSystemMetrics( SM_CXSCREEN ) - g_DISPLAY_iLCDBackground_Width ) / 2;
    iWindowY = ( GetSystemMetrics( SM_CYSCREEN ) - g_DISPLAY_iLCDBackground_Height ) / 2;

    // create window
    g_DISPLAY_hWnd = CreateWindow( DISPLAY_WINDOW_CLASSNAME, DISPLAY_WINDOW_TITLE, WS_POPUP, iWindowX, iWindowY, g_DISPLAY_iLCDBackground_Width, g_DISPLAY_iLCDBackground_Height, NULL, NULL, hInstance, NULL );
    CHECK( g_DISPLAY_hWnd == NULL, CMG_ERROR_OS )

    // show and update window
    ShowWindow( g_DISPLAY_hWnd, SW_SHOW );
    UpdateWindow( g_DISPLAY_hWnd );
    // bring new window to foreground
    SetForegroundWindow ( g_DISPLAY_hWnd );



    // . . . . . . . . . . . . . . . . . . . . . . .
    // create popup menu
    g_DISPLAY_hMenu = CreatePopupMenu();
    CHECK( g_DISPLAY_hMenu == NULL, CMG_ERROR_OS );
    AppendMenu( g_DISPLAY_hMenu, MF_STRING, MENU_MODE_PHOTO, MENU_MODE_PHOTO_TEXT );
    AppendMenu( g_DISPLAY_hMenu, MF_STRING, MENU_MODE_EXACT, MENU_MODE_EXACT_TEXT );
    


    // . . . . . . . . . . . . . . . . . . . . . . .
    // create a timer to update the content
    g_DISPLAY_uTimerID = SetTimer( g_DISPLAY_hWnd, 1, DISPLAY_WINDOW_TIMER_MS, NULL );
    CHECK( g_DISPLAY_uTimerID == 0, CMG_ERROR_OS )



    // . . . . . . . . . . . . . . . . . . . . . . .
    // signalize the event
    // tell caller thread that we are ready now
    CHECK( SetEvent( hWaitEvent ) == 0, CMG_ERROR_OS )



    // . . . . . . . . . . . . . . . . . . . . . . .
    // message loop
    while ( GetMessage( &msg, NULL, 0, 0 ) )
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }



    // . . . . . . . . . . . . . . . . . . . . . . .
    // cleanup, free and destroy
    DestroyMenu( g_DISPLAY_hMenu );

    // reset
    g_DISPLAY_hWnd = NULL;
    g_DISPLAY_hThread = NULL;

    return CMG_OK;


    //.............................
    // on error
_Error:
    // cleanup, free and destroy
    if ( g_DISPLAY_hMenu )
        DestroyMenu( g_DISPLAY_hMenu );

    // reset
    g_DISPLAY_hWnd = NULL;
    g_DISPLAY_hThread = NULL;
    return iResult;
}




/*---------------------------------------------------------------------------
| _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer
|-----------------------------------
| redraws all pixels from buffer
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    none
+--------------------------------------------------------------------------*/
_PRIVATE
void _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer( void )
{
    cmg_Coord   iX, iY;
    cmg_u8      *pbyLCDBuffer;      // local copy

    // load pixel color color
    if ( g_DISPLAY_bExactMode )
    {
        CMG_LLIO_DISPLAY_COMMON_CreatePixelColor_2x2_32bpp( g_DISPLAY_aadwPixelColor, CMG_LLIO_DISPLAY_COMMON_PIXELCOLORTYPE_EXACT );
    }
    else 
    {
        if ( g_DISPLAY_bBacklightOn )
            CMG_LLIO_DISPLAY_COMMON_CreatePixelColor_2x2_32bpp( g_DISPLAY_aadwPixelColor, CMG_LLIO_DISPLAY_COMMON_PIXELCOLORTYPE_PHOTO_BLON );
        else
            CMG_LLIO_DISPLAY_COMMON_CreatePixelColor_2x2_32bpp( g_DISPLAY_aadwPixelColor, CMG_LLIO_DISPLAY_COMMON_PIXELCOLORTYPE_PHOTO_BLOFF );
    }

    // prepare lcd bitmap (clear all pixels)
    pbyLCDBuffer = g_DISPLAY_pbyLCDBuffer;     // local copy
    for ( iY = 0; iY < g_DISPLAY_iHeight; iY++ )
        for ( iX = 0; iX < g_DISPLAY_iWidth; iX++ )
            CMG_LLIO_DISPLAY_SetPixel( iX, iY, *pbyLCDBuffer++ );
}

//---------------------------------------------------------------------------





//---------------------------------------------------------------------------
// exportet functions
//---------------------------------------------------------------------------


/****************************************************************************
| CMG_LLIO_DISPLAY_Init
|-----------------------------------
| Init (init downwards: nothing)
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    result
+***************************************************************************/
_PUBLIC
cmg_Result CMG_LLIO_DISPLAY_Init( void )
{
    cmg_Result  iResult;
    HANDLE      hWaitEvent = NULL;


    // . . . . . . . . . . . . . . . . . . . . . . .
    // init stuff

    // set the display resolution
    {
        cmg_mu32    uIdx;

        g_DISPLAY_iWidth = CMG_LLIO_EMU_WIDTH;
        g_DISPLAY_iHeight = CMG_LLIO_EMU_HEIGHT;
        g_DISPLAY_bLCD_On = true;
        g_DISPLAY_bContentChanged = false;
        g_DISPLAY_bModeChanged = false;
        g_DISPLAY_dwErrors = 0;
        for ( uIdx = 0; uIdx < MAX_NUMBER_OF_ERRORS; uIdx++ )
            g_DISPLAY_astrError[ uIdx ] = NULL;
    }


    // set byte order
    CMG_LLIO_DISPLAY_COMMON_SetByteOrder( CMG_LLIO_DISPLAY_COMMON_BYTEORDER_GDI );



    // . . . . . . . . . . . . . . . . . . . . . . .
    // lcd buffer
    // allocate lcd buffer (for redraw in another color mode)
    g_DISPLAY_pbyLCDBuffer = malloc( g_DISPLAY_iWidth * g_DISPLAY_iHeight );
    CHECK( g_DISPLAY_pbyLCDBuffer == NULL, CMG_ERROR_NOMEM );

    // prepare buffer (clear all pixels)
    memset( g_DISPLAY_pbyLCDBuffer, 0, g_DISPLAY_iWidth * g_DISPLAY_iHeight );

    

    // . . . . . . . . . . . . . . . . . . . . . . .
    // lcd image
    // each emulated lcd pixel is 2x2 pixels large
    g_DISPLAY_iLCD_Width = g_DISPLAY_iWidth * 2;
    g_DISPLAY_iLCD_Height = g_DISPLAY_iHeight * 2;
    
    // allocate lcd bitmap (each pixels needs 4 bytes)
    g_DISPLAY_pdwLCD = malloc( g_DISPLAY_iLCD_Width * g_DISPLAY_iLCD_Height * sizeof(cmg_u32) );
    CHECK( g_DISPLAY_pdwLCD == NULL, CMG_ERROR_NOMEM );

    // set the bitmap info for the lcd bitmap
    memset( &g_DISPLAY_pLCD_BI, 0, sizeof(BITMAPINFO) );
    g_DISPLAY_pLCD_BI.bmiHeader.biSize = sizeof( BITMAPINFOHEADER );    // structure size
    g_DISPLAY_pLCD_BI.bmiHeader.biWidth = g_DISPLAY_iLCD_Width;         // width
    g_DISPLAY_pLCD_BI.bmiHeader.biHeight = -g_DISPLAY_iLCD_Height;      // height (inverted for top-down bitmap)
    g_DISPLAY_pLCD_BI.bmiHeader.biPlanes = 1;                           // fixed 1
    g_DISPLAY_pLCD_BI.bmiHeader.biBitCount = 32;                        // 32 bits per pixel
    g_DISPLAY_pLCD_BI.bmiHeader.biCompression = BI_RGB;                 // plain RGB


    
    // . . . . . . . . . . . . . . . . . . . . . . .
    // create pixel color table and prepare lcd bitmap
    // redraw from buffer (=> prepare lcd bitmap (clear all pixels))
    g_DISPLAY_bBacklightOn = true;
    g_DISPLAY_bExactMode = false;
    _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer();



    // . . . . . . . . . . . . . . . . . . . . . . .
    // lcd background image
    // create image
    g_DISPLAY_iLCDBackground_Width = g_DISPLAY_iLCD_Width + ( 2 * CMG_LLIO_DISPLAY_COMMON_LCDBG_START_X );
    g_DISPLAY_iLCDBackground_Height = g_DISPLAY_iLCD_Height + ( 2 * CMG_LLIO_DISPLAY_COMMON_LCDBG_START_Y );

    iResult = CMG_LLIO_DISPLAY_COMMON_CreateLCDImage32( &g_DISPLAY_pdwLCDBackground, g_DISPLAY_iLCDBackground_Width, g_DISPLAY_iLCDBackground_Height );
    CHECKRESULT();
    
    // set the bitmap info for the lcd background bitmap
    memset( &g_DISPLAY_pLCDBackground_BI, 0, sizeof(BITMAPINFO) );
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biSize = sizeof( BITMAPINFOHEADER );      // structure size
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biWidth = g_DISPLAY_iLCDBackground_Width; // width
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biHeight = -g_DISPLAY_iLCDBackground_Height; // height (inverted for top-down bitmap)
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biPlanes = 1;                             // fixed 1
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biBitCount = 32;                          // 32 bits per pixel
    g_DISPLAY_pLCDBackground_BI.bmiHeader.biCompression = BI_RGB;                   // plain RGB


    
    // . . . . . . . . . . . . . . . . . . . . . . .
    // create and start thread
    {
        DWORD       dwThreadId;
        DWORD       wfsoResult;

        // create event to wait until thread is ready (window is correctly created)
        hWaitEvent = CreateEvent( NULL, 0, 0, NULL );
        CHECK( hWaitEvent == NULL, CMG_ERROR_OS );

        // create and start thread and submit hWaitEvent as parameter
        g_DISPLAY_hThread = CreateThread( NULL, 0, _DISPLAY_ThreadProc, &hWaitEvent, 0, &dwThreadId );
        CHECK( g_DISPLAY_hThread == NULL, CMG_ERROR_OS );

        // wait for event (window is correctly created)
        wfsoResult = WaitForSingleObject( hWaitEvent, DISPLAY_WINDOW_WAITTIME_MS );
        CHECK( ( wfsoResult == WAIT_FAILED ) || ( wfsoResult == WAIT_TIMEOUT ), CMG_ERROR_OS );
        CHECK( ( g_DISPLAY_hThread == NULL ) || ( g_DISPLAY_hWnd == NULL ), CMG_ERROR_OS );

    
        // close and free event
        CloseHandle( hWaitEvent );
    }
    
    // return success
    return CMG_OK;


    //.............................
    // on error free everything
_Error:
    // close and free event
    if ( hWaitEvent )
        CloseHandle( hWaitEvent );

    CMG_LLIO_DISPLAY_Exit();
    return iResult;
}



/****************************************************************************
| CMG_LLIO_DISPLAY_Exit
|-----------------------------------
| Exit (exit downwards: nothing)
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_DISPLAY_Exit( void )
{
    cmg_mu32    uIdx;

    // . . . . . . . . . . . . . . . . . . . . . . .
    // shutdown lcd window
    // first destroy window
    if ( g_DISPLAY_hWnd )
        SendMessage( g_DISPLAY_hWnd, WM_DESTROY, 0, 0 );

    // if thread is running wait for it (with timeout)
    if ( g_DISPLAY_hThread )
        WaitForSingleObject( g_DISPLAY_hThread, DISPLAY_WINDOW_WAITTIME_MS );

    // if thread is STILL running kill it
    if ( g_DISPLAY_hThread )
        TerminateThread( g_DISPLAY_hThread, CMG_ERROR_OS );


    // . . . . . . . . . . . . . . . . . . . . . . .
    // free memory
    SFREE( g_DISPLAY_pbyLCDBuffer );
    SFREE( g_DISPLAY_pdwLCD );
    SFREE( g_DISPLAY_pdwLCDBackground );    
    for ( uIdx = 0; uIdx < MAX_NUMBER_OF_ERRORS; uIdx++ )
        SFREE( g_DISPLAY_astrError[ uIdx ] );
}



/****************************************************************************
| CMG_LLIO_DISPLAY_PowerManagement
|-----------------------------------
| Power Management
+----------------------------------------------------------------------------
| @Params:         
|    uState:        switch to powerstate
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_DISPLAY_PowerManagement( cmg_mu8 uState )
{
    switch ( uState )
    {
        // display off
        case CMG_PWR_DISPLAY_OFF:
            // ... nothing to do ...
            // the driver layer will turn off the display
            break;


        // display on
        case CMG_PWR_DISPLAY_ON:
            // ... nothing to do ...
            // the driver layer will turn on the display
            break;


        // backlight on
        case CMG_PWR_BACKLIGHT_ON:
            g_DISPLAY_bBacklightOn = true;
            _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer();
            break;


        // backlight off
        case CMG_PWR_BACKLIGHT_OFF:
            g_DISPLAY_bBacklightOn = false;
            _DISPLAY_LoadAndApplyPixelColorTableAndRedrawFromBuffer();
            break;
    }
}



/****************************************************************************
| CMG_LLIO_DISPLAY_SetPixel
|-----------------------------------
| Set a new pixel color
+----------------------------------------------------------------------------
| @Params:         
|   * iX:           x coordinate
|   * iY:           y coordinate
|   * byColor:      grayscale color
| @Return Value:    none
+***************************************************************************/
void CMG_LLIO_DISPLAY_SetPixel( cmg_Coord iX, cmg_Coord iY, cmg_Color Color )
{
    // check range
    if ( ( iX >= 0 ) && ( iX < g_DISPLAY_iWidth ) && ( iY >= 0 ) && ( iY < g_DISPLAY_iHeight ) )
    {
        // first set in internal buffer
        g_DISPLAY_pbyLCDBuffer[ ( iY * g_DISPLAY_iWidth  ) + iX ] = Color;

        // set the 2x2 coresponding pixels
        g_DISPLAY_pdwLCD[ ( ( iY<<1 ) + 0 ) * g_DISPLAY_iLCD_Width + ( iX<<1 ) + 0 ] = g_DISPLAY_aadwPixelColor[ Color ][ 0 ];
        g_DISPLAY_pdwLCD[ ( ( iY<<1 ) + 0 ) * g_DISPLAY_iLCD_Width + ( iX<<1 ) + 1 ] = g_DISPLAY_aadwPixelColor[ Color ][ 1 ];
        g_DISPLAY_pdwLCD[ ( ( iY<<1 ) + 1 ) * g_DISPLAY_iLCD_Width + ( iX<<1 ) + 0 ] = g_DISPLAY_aadwPixelColor[ Color ][ 2 ];
        g_DISPLAY_pdwLCD[ ( ( iY<<1 ) + 1 ) * g_DISPLAY_iLCD_Width + ( iX<<1 ) + 1 ] = g_DISPLAY_aadwPixelColor[ Color ][ 3 ];

        // signalize changes
        g_DISPLAY_bContentChanged = true;
    }
}



/****************************************************************************
| CMG_LLIO_DISPLAY_SetMode
|-----------------------------------
| Set the mode
+----------------------------------------------------------------------------
| @Params:          byMode: defined in .h file
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void    CMG_LLIO_DISPLAY_SetMode( cmg_mu8 uMode )
{
    // select mode
    switch ( uMode )
    {
        // turn display off
        case CMG_LLIO_DISPLAY_MODE_OFF:
            g_DISPLAY_bLCD_On = false;
            break;

        // turn display on
        case CMG_LLIO_DISPLAY_MODE_ON:
            g_DISPLAY_bLCD_On = true;
            break;
    }
    
    // set mode changed flag   
    g_DISPLAY_bModeChanged = true;
}




/****************************************************************************
| CMG_LLIO_DISPLAY_ErrorMessage
|-----------------------------------
| Display an error message
+----------------------------------------------------------------------------
| @Params:         
|   * strMessage:   message to display
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_DISPLAY_ErrorMessage( cmg_char *strMessage )
{
    // if there is enough space, add the error
    if ( g_DISPLAY_dwErrors < MAX_NUMBER_OF_ERRORS )
    {
        // alloc mem for error
        g_DISPLAY_astrError[ g_DISPLAY_dwErrors ] = malloc( strlen( strMessage ) + 1 );
        // copy message
        strcpy( g_DISPLAY_astrError[ g_DISPLAY_dwErrors ], strMessage );
        // increment number of errors
        g_DISPLAY_dwErrors++;
    }
}

//---------------------------------------------------------------------------



#endif

/////////////////////////////////////////////////////////////////////////////
// (c) 2oo6 - 2oo8 Christian Merkle
