//***************************************************************************
// CMG_LLIO_EMU_EMULATOR_T6963.c
//***************************************************************************
// CMG - Christian Merkle Graphics
// Copyright (C) 2oo6 - 2oo8 Christian Merkle 
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// To contact Christian Merkle, see his website at:
// http://www.cmerkle.de
//***************************************************************************
// Author:          Christian Merkle
// Created:         30.10.2006
//***************************************************************************
// History:
//   04.11.2006     *stable*
//   23.01.2007     added power management
//***************************************************************************
// ToDo:
//   [ ]    xxx
//***************************************************************************



// main CMG include (will include user configuration)
#include    "../../CMG_Main.h"



//---------------------------------------------------------------------------
// conditional compilation and configuration checks
//---------------------------------------------------------------------------
#ifndef CMG_LLIO_CONFIGURED
    #error CMG_LLIO not configured. Please customize the config file correctly (use the *.config template).
#elif defined( CMG_LLIO_EMU_EMULATOR_T6963 )

    #if CMG_LLIO_EMU_BITDEPTH != 1
        #error CMG_LLIO_EMU_EMULATOR_T6963 will only work with CMG_LLIO_EMU_BITDEPTH == 1
    #endif
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// inludes
//---------------------------------------------------------------------------
// CMG:
#include    "../CMG_LLIO.h"
#include    "CMG_LLIO_EMU_DISPLAY.h"
// other:
#include    <stdio.h>
#include    <malloc.h>
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// defines
//---------------------------------------------------------------------------

// address line defines
#define     CMG_LLIO_WriteCommand           CMG_LLIO_WriteA1
#define     CMG_LLIO_WriteData              CMG_LLIO_WriteA0
#define     CMG_LLIO_ReadStatus             CMG_LLIO_ReadA1
#define     CMG_LLIO_ReadData               CMG_LLIO_ReadA0


// T6963 commands   (* marks handled, ~ marks handled partially)
#define     CMD_CURSOR_PTR                  0x21    // 
#define     CMD_OFFSET_REG                  0x22    // 
#define     CMD_ADDRESS_PTR                 0x24    // *

#define     CMD_TEXT_HOME_ADDR              0x40    // 
#define     CMD_TEXT_AREA                   0x41    // 
#define     CMD_GRAPHIC_HOME_ADDR           0x42    // ~
#define     CMD_GRAPHIC_AREA                0x43    // *

#define     CMD_MODE_OR                     0x80    // *
#define     CMD_MODE_XOR                    0x81    // 
#define     CMD_MODE_AND                    0x83    // 
#define     CMD_MODE_TEXT_ATTRIB            0x84    // 

#define     CMD_DISPLAY_MODE_OFF            0x90    // *
#define     CMD_DISPLAY_MODE_GRAPHIC_ONLY   0x98    // *

#define     CMD_CURSOR_PATTERN_SELECT       0xa0    // 

#define     CMD_DATA_AUTO_WRITE             0xb0    // *
#define     CMD_DATA_AUTO_READ              0xb1    // *
#define     CMD_DATA_AUTO_RESET             0xb2    // *

#define     CMD_DATA_WRITE_INC              0xc0    // *
#define     CMD_DATA_READ_INC               0xc1    // *
#define     CMD_DATA_WRITE_DEC              0xc2    // *
#define     CMD_DATA_READ_DEC               0xc3    // *
#define     CMD_DATA_WRITE                  0xc4    // *
#define     CMD_DATA_READ                   0xc5    // *

#define     CMD_BIT_SETCLR                  0xf0    // *


// status bits
#define     WAIT_COMMAND_DATA_BITS          0x03
#define     WAIT_AUTOMODE_READ_BITS         0x04
#define     WAIT_AUTOMODE_WRITE_BITS        0x08


// error messages
#define     STR_COMMAND_UNKOWN              "Unknown or unsupported command: CMD=0x%02x D1/2=0x%02x D2/1=0x%02x"


// maximum controller memory size
#define     T6963_MEMORY_SIZE               64 * KB

#define     COLOR_OFF                       0x00
#define     COLOR_ON                        0xff
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// globals
//---------------------------------------------------------------------------
cmg_u8      g_EMULATOR_byD1;                    // data 1 register
cmg_u8      g_EMULATOR_byD2;                    // data 2 register
cmg_u8      g_EMULATOR_byDRet;                  // data return register

cmg_u16     g_EMULATOR_wAddressPtr;             // current address pointer
cmg_u16     g_EMULATOR_wGraphicHomeAddr;        // graphic home address
cmg_u8      g_EMULATOR_byGraphicArea;           // graphic area (linesize in bytes)
cmg_u8      g_EMULATOR_byAutoMode;              // current automode

cmg_u8      *g_EMULATOR_pbyVideoMem = NULL;     // emulator native video memory
//---------------------------------------------------------------------------





//---------------------------------------------------------------------------
// functions
//---------------------------------------------------------------------------



/****************************************************************************
| CMG_LLIO_Init
|-----------------------------------
| Init (init downwards: CMG_LLIO_DISPLAY_Init)
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    result
+***************************************************************************/
_PUBLIC
cmg_Result CMG_LLIO_Init( void )
{
    cmg_Result  iResult;

    // init CMG_LLIO_DISPLAY_*
    iResult = CMG_LLIO_DISPLAY_Init();
    CHECKRESULT();
    
    // reset vars
    g_EMULATOR_byD1 = g_EMULATOR_byD2 = g_EMULATOR_byDRet = 0;
    g_EMULATOR_wAddressPtr = g_EMULATOR_wGraphicHomeAddr = 0;
    g_EMULATOR_byGraphicArea = 1;     // because of divide by zero possibility

    // automode off
    g_EMULATOR_byAutoMode = 0;

    // allocate video memory
    g_EMULATOR_pbyVideoMem = malloc( T6963_MEMORY_SIZE );
    CHECK( g_EMULATOR_pbyVideoMem == NULL, CMG_ERROR_NOMEM );

    // return success
    return CMG_OK;

    //.............................
    // on error free everything
_Error:
    CMG_LLIO_Exit();
    return iResult;
}



/****************************************************************************
| CMG_LLIO_Exit
|-----------------------------------
| Exit (exit downwards: CMG_LLIO_DISPLAY_Exit)
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_Exit( void )
{
    // free video memory
    if ( g_EMULATOR_pbyVideoMem )
    {
        free( g_EMULATOR_pbyVideoMem );
        g_EMULATOR_pbyVideoMem = NULL;
    }

    // exit CMG_LLIO_DISPLAY_*
    CMG_LLIO_DISPLAY_Exit();
}



/****************************************************************************
| CMG_LLIO_PowerManagement
|-----------------------------------
| Power Management
+----------------------------------------------------------------------------
| @Params:         
|    uState:        switch to powerstate
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_PowerManagement( cmg_mu8 uState )
{
    switch ( uState )
    {
        // display off
        case CMG_PWR_DISPLAY_OFF:
            // ... nothing to do ...

            // on off call parent at last
            CMG_LLIO_DISPLAY_PowerManagement( uState );
            break;


        // display on
        case CMG_PWR_DISPLAY_ON:
            // on on call parent first
            CMG_LLIO_DISPLAY_PowerManagement( uState );

            // ... nothing to do ...
            break;


        // just pass trough unknown
        default:
            // case CMG_PWR_BACKLIGHT_ON:
            // case CMG_PWR_BACKLIGHT_OFF:
            CMG_LLIO_DISPLAY_PowerManagement( uState );
            break;
    }
}





/*---------------------------------------------------------------------------
| _LLIO_Error_UnknownCommand
|-----------------------------------
| Sends an UnknownCommand message to the display emulator
+----------------------------------------------------------------------------
| @Params:          command byte
| @Return Value:    none
+--------------------------------------------------------------------------*/
_PRIVATE
void _LLIO_Error_UnknownCommand( cmg_u8 byCommand )
{
    cmg_char    strMessage[ sizeof(STR_COMMAND_UNKOWN) + 12 ];
    sprintf( strMessage, STR_COMMAND_UNKOWN, byCommand, g_EMULATOR_byD1, g_EMULATOR_byD2 );
    CMG_LLIO_DISPLAY_ErrorMessage( strMessage );
}



/*---------------------------------------------------------------------------
| _LLIO_SendCurrentBitToDisplay
|-----------------------------------
| 
+----------------------------------------------------------------------------
| @Params:          bit number
| @Return Value:    none
+--------------------------------------------------------------------------*/
_PRIVATE
void _LLIO_SendCurrentBitToDisplay( cmg_u8 byBit )
{
    cmg_Coord   iX, iY;
    cmg_u8      byData;
    
    // calculate x and y from address pointer
    iX = ( ( g_EMULATOR_wAddressPtr % g_EMULATOR_byGraphicArea ) << 3 ) + ( 7 - byBit );
    iY = g_EMULATOR_wAddressPtr / g_EMULATOR_byGraphicArea;

    // get data byte
    byData = g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ];
    
    // set pixel
    CMG_LLIO_DISPLAY_SetPixel( iX, iY, (cmg_u8)( ( byData & BIT( byBit ) ) ? COLOR_ON : COLOR_OFF ) );
}



/*---------------------------------------------------------------------------
| _LLIO_SendCurrentByteToDisplay
|-----------------------------------
| 
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    none
+--------------------------------------------------------------------------*/
_PRIVATE
void _LLIO_SendCurrentByteToDisplay( void )
{
    cmg_Coord   iX, iY;
    cmg_u8      byData;
    cmg_u8      byBit;
    cmg_u8      byBitMask;
    
    // calculate x and y from address pointer
    iX = ( g_EMULATOR_wAddressPtr % g_EMULATOR_byGraphicArea ) << 3;
    iY = g_EMULATOR_wAddressPtr / g_EMULATOR_byGraphicArea;

    // get data byte
    byData = g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ];
    
    // start with leftmost bit
    byBitMask = BIT( 7 );
    for ( byBit = 0; byBit < 8; byBit++ )
    {
        // set pixel
        CMG_LLIO_DISPLAY_SetPixel( (cmg_Coord)( iX + byBit ), iY, (cmg_u8)( ( byData & byBitMask ) ? COLOR_ON : COLOR_OFF ) );
        // continue with next right bit
        byBitMask >>= 1;
    }
}



/****************************************************************************
| CMG_LLIO_WriteCommand
|-----------------------------------
| Write byte in Command Register
+----------------------------------------------------------------------------
| @Params:          command byte
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_WriteCommand( cmg_u8 byCommand )
{
    switch ( byCommand )
    {
        // register set commands:
        case CMD_ADDRESS_PTR:
            // set address pointer
            g_EMULATOR_wAddressPtr = BUILD_WORD( g_EMULATOR_byD1, g_EMULATOR_byD2 );
            break;

        case CMD_GRAPHIC_HOME_ADDR:
            // set address pointer
            g_EMULATOR_wGraphicHomeAddr = BUILD_WORD( g_EMULATOR_byD1, g_EMULATOR_byD2 );
            if ( g_EMULATOR_wGraphicHomeAddr != 0 )
            {
                _LLIO_Error_UnknownCommand( byCommand );
                return;
            }
            break;

        case CMD_GRAPHIC_AREA:
            // D1 must be 0
            if ( g_EMULATOR_byD1 != 0 )
            {
                _LLIO_Error_UnknownCommand( byCommand );
                return;
            }
            // set graphic area
            g_EMULATOR_byGraphicArea = g_EMULATOR_byD2;
            break;
            

        // mode commands:
        case CMD_MODE_OR:
            // asume always mode or (every other mode will result in an unknown command error)
            break;


        // display commands:
        case CMD_DISPLAY_MODE_OFF:
            // switch display off
            CMG_LLIO_DISPLAY_SetMode( CMG_LLIO_DISPLAY_MODE_OFF );
            break;

        case CMD_DISPLAY_MODE_GRAPHIC_ONLY:
            // switch display on
            CMG_LLIO_DISPLAY_SetMode( CMG_LLIO_DISPLAY_MODE_ON );
            break;


        // auto mode commands:
        case CMD_DATA_AUTO_WRITE:
            // enable auto write
            g_EMULATOR_byAutoMode = WAIT_AUTOMODE_WRITE_BITS;
            break;

        case CMD_DATA_AUTO_READ:
            // enable auto read
            g_EMULATOR_byAutoMode = WAIT_AUTOMODE_READ_BITS;
            break;

        case CMD_DATA_AUTO_RESET:
            // enable auto mode
            g_EMULATOR_byAutoMode = 0;
            break;


        // single write/read commands:
        case CMD_DATA_WRITE_INC:
            // write inc
            g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ] = g_EMULATOR_byD1;
            _LLIO_SendCurrentByteToDisplay();
            g_EMULATOR_wAddressPtr++;
            break;

        case CMD_DATA_READ_INC:
            // read inc
            g_EMULATOR_byDRet = g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr++ ];
            break;

        case CMD_DATA_WRITE_DEC:
            // write dec
            g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ] = g_EMULATOR_byD1;
            _LLIO_SendCurrentByteToDisplay();
            g_EMULATOR_wAddressPtr--;
            break;

        case CMD_DATA_READ_DEC:
            // read dec
            g_EMULATOR_byDRet = g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr-- ];
            break;

        case CMD_DATA_WRITE:
            // write
            g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ] = g_EMULATOR_byD1;
            _LLIO_SendCurrentByteToDisplay();
            break;

        case CMD_DATA_READ:
            // read
            g_EMULATOR_byDRet = g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ];
            break;



        default:
            // bit set/clear commands:
            if ( ( byCommand & CMD_BIT_SETCLR ) == CMD_BIT_SETCLR )
            {
                cmg_u8  byBit = byCommand & 0x07;

                if ( byCommand & 0x08 )
                {
                    // set bit
                    SET_BIT( g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ], byBit );
                }
                else
                {
                    // clr bit
                    CLR_BIT( g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ], byBit );
                }

                _LLIO_SendCurrentBitToDisplay( byBit );
                break;
            }


            // everything else:
            _LLIO_Error_UnknownCommand( byCommand );
            break;
    }
}



/****************************************************************************
| CMG_LLIO_WriteData
|-----------------------------------
| Write byte in Data Register
+----------------------------------------------------------------------------
| @Params:          data byte
| @Return Value:    none
+***************************************************************************/
_PUBLIC
void CMG_LLIO_WriteData( cmg_u8 byData )
{
    // automode?
    if ( g_EMULATOR_byAutoMode == WAIT_AUTOMODE_WRITE_BITS )
    {
        // automode active -> write data to display mem
        g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr ] = byData;
        _LLIO_SendCurrentByteToDisplay();
        g_EMULATOR_wAddressPtr++;
        return;
    }

    // normal mode -> write data to data registers
    g_EMULATOR_byD2 = g_EMULATOR_byD1;
    g_EMULATOR_byD1 = byData;
}



/****************************************************************************
| CMG_LLIO_ReadStatus
|-----------------------------------
| Read byte from Status Register
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    status byte
+***************************************************************************/
_PUBLIC
cmg_u8 CMG_LLIO_ReadStatus( void )
{
    // return current status
    return WAIT_COMMAND_DATA_BITS | g_EMULATOR_byAutoMode;
}



/****************************************************************************
| CMG_LLIO_ReadData
|-----------------------------------
| Read byte from Data Register
+----------------------------------------------------------------------------
| @Params:          none
| @Return Value:    data byte
+***************************************************************************/
_PUBLIC 
cmg_u8 CMG_LLIO_ReadData( void )
{
    // automode active -> read data from display mem
    if ( g_EMULATOR_byAutoMode == WAIT_AUTOMODE_READ_BITS )
        return g_EMULATOR_pbyVideoMem[ g_EMULATOR_wAddressPtr++ ];

    // normal mode -> return return register
    return g_EMULATOR_byDRet;
}
//---------------------------------------------------------------------------


#endif

/////////////////////////////////////////////////////////////////////////////
// (c) 2oo6 - 2oo8 Christian Merkle
