//------------------------------------------------------------------------------------------------- 
// Copyright (c) 2oo7, 2oo8 CmDev, Christian Merkle. All rights reserved.       http://www.cmdev.de
//-------------------------------------------------------------------------------------------------
// Author of this file: christian@cmdev.de

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace FontImport
{
    class CMGFont
    {
        private string _fontName;
        public string FontName
        { get { return _fontName; } }
        
        private Font _windowsFont;
        public Font WindowsFont
        { get { return _windowsFont; } }

        private int _fontType = 100;

        private int _startChar;
        private int _stopChar;

        private Size _maxCharSize;
        public Size MaxCharSize
        { get { return _maxCharSize; } }


        public int LineSizeBytes
        { get { return ( ( ( _maxCharSize.Width - 1 ) >> 3 ) + 1 ); } }

        private int CharSizeBytes
        { get { return LineSizeBytes * _maxCharSize.Height; } }

        private int CharOffsetSizeBytes
        { get { return 1 + CharSizeBytes; } }

        private int FontHeaderBytes
        { get { return 5; } }

        private int FontSizeBytes
        { get { return ( CharOffsetSizeBytes * _chars.Count ) + FontHeaderBytes; } }


        private List<CMGFontChar> _chars = new List<CMGFontChar>();



        static public string GetStringFromRange( int startChar, int stopChar )
        {
            StringBuilder str = new StringBuilder();
            
            for ( int i = startChar; i <= stopChar; i++ )
                str.Append( Convert.ToChar( i ) );
            
            return str.ToString();
        }




        public CMGFont( Font windowsFont, int startChar, int stopChar, bool showWaitForm )
        {
            WaitForm waitForm = null;

            _fontName = windowsFont.Name + " " + windowsFont.Size.ToString() + ( windowsFont.Bold ? " Bold" : "" ) + ( windowsFont.Italic ? " Italic" : "" );
            _windowsFont = windowsFont;
            _startChar = startChar;
            _stopChar = stopChar;


            if ( showWaitForm )
            {
                waitForm = new WaitForm();
                waitForm.SetBounds( _startChar, _stopChar );
                waitForm.Show();
            }

            

            string characters = GetStringFromRange( startChar, stopChar );

            // mesasure each char to find max size
            _maxCharSize = new Size();
            foreach ( char c in characters )
            {
                Size size = CMGFontChar.MeasureChar( c, windowsFont );
                
                if ( _maxCharSize.Width < size.Width )
                    _maxCharSize.Width = size.Width;
                
                if ( _maxCharSize.Height < size.Height )
                    _maxCharSize.Height = size.Height;
            }

            // load each char
            foreach ( char c in characters )
            {
                if ( showWaitForm )
                    waitForm.IncValue();

                _chars.Add( new CMGFontChar( this, c ) );
            }

            if ( showWaitForm )
                waitForm.Close();
        }



                         
        public void Optimize()
        {
            // remove unused lines
            while ( _OptimizeLine( 0 ) );
            while ( _OptimizeLine( _maxCharSize.Height - 1 ) );

            // optimize all char widths
            foreach ( CMGFontChar charData in _chars )
                charData.Optimize();

            // search new max width
            _maxCharSize.Width = 0;
            foreach ( CMGFontChar charData in _chars )
                if ( _maxCharSize.Width < charData.Size.Width )
                    _maxCharSize.Width = charData.Size.Width;
        }


        private bool _OptimizeLine( int line )
        {
            // not legal line
            if ( ( line < 0 ) || ( line >= _maxCharSize.Height ) )
                return false;

            // search all chars if line is empty
            foreach ( CMGFontChar charData in _chars )
                foreach ( bool pixel in charData.PixelData[ line ] )
                    if ( pixel )
                        return false;

            // if we are here we can safely delete the line
            foreach ( CMGFontChar charData in _chars )
            {
                charData.Size.Height--;
                charData.PixelData.RemoveAt( line );
            }

            _maxCharSize.Height--;

            return true;
        }


        public string CompileToCCode()
        {
            StringBuilder code = new StringBuilder();

            code.AppendLine( "/****************************************************************\\");
            code.AppendFormat( "| CMG Font: {0}", FontName ); 
            code.AppendLine();
            code.AppendLine( "| Font created by CMGFontConvert (c) 2oo7, 2oo8 Christian Merkle" );
            code.AppendLine( "| Please respect the copyright of the original font." );
            code.AppendLine( "|-----------------------------------------------------------------");
            code.AppendLine( "| Font Type: uncompressed" );
            code.AppendFormat( "| Start char: {0}, stop char: {1}, number of chars: {2}", _startChar, _stopChar, _chars.Count );
            code.AppendLine();
            code.AppendFormat( "| Max char size: {0}", _maxCharSize );
            code.AppendLine();
            code.AppendFormat( "| Line size: {0} bytes", LineSizeBytes );
            code.AppendLine();
            code.AppendFormat( "| Char size: {0} bytes", CharOffsetSizeBytes );
            code.AppendLine();
            code.AppendFormat( "| Header size: {0} bytes", FontHeaderBytes );
            code.AppendLine();
            code.AppendFormat( "| Font size: {0} bytes", FontSizeBytes );
            code.AppendLine();
            code.AppendLine( "\\****************************************************************/");
            code.AppendLine();
            code.AppendLine( "#include \"CMG/CMG.h\"" );
            code.AppendLine();
            code.AppendFormat( "cmg_FontRawData    g_Font_{0}[] = ", FontName.Replace( ' ', '_' ) );
            code.AppendLine();
            code.AppendLine( "    {" );

            string lineStart = new string( ' ', 8 );

            code.AppendLine( lineStart + "/* Font Header (type, start, stop, linesize, height): */" );
            code.AppendFormat( "{0}{1}, {2}, {3}, {4}, {5},", lineStart, _ValToC( _fontType ), _ValToC( _startChar ), _ValToC( _stopChar ), _ValToC( LineSizeBytes ), _ValToC( _maxCharSize.Height ) );

            foreach ( CMGFontChar charData in _chars )
            {
                code.AppendLine();
                code.AppendLine();

                code.AppendFormat( "{0}/* Character Data \"{1}\" (width, {2} data bytes): */", lineStart, charData.Char, CharSizeBytes );
                code.AppendLine();
                
                code.Append( lineStart );
                foreach ( byte val in charData.CompileChar() )
                    code.AppendFormat( "{0}, ", _ValToC( val ) );
            }

            code.Remove( code.Length - 2, 2 );
            code.AppendLine();
            code.AppendLine( "    };" );

            return code.ToString();
        }


        private string _ValToC( int val )
        {
            return string.Format( "0x{0}", val.ToString( "x2" ) );
        }


        
        public override string ToString()
        {
            return string.Format( "Font Name: {9}{0}Font Type: uncompressed{0}Start char: {1}, Stop char: {2}, Number of chars: {3}{0}Max char size: {4}{0}Line size: {5} Bytes{0}Char size: {6} Bytes{0}Header size: {7} Bytes{0}Font Size: {8} Bytes",
                Environment.NewLine,
                _startChar, _stopChar, _chars.Count,
                _maxCharSize,
                LineSizeBytes,
                CharOffsetSizeBytes,
                FontHeaderBytes, 
                FontSizeBytes,
                FontName
                );
        }
    }
}
