//------------------------------------------------------------------------------------------------- 
// Copyright (c) 2oo7, 2oo8 CmDev, Christian Merkle. All rights reserved.       http://www.cmdev.de
//-------------------------------------------------------------------------------------------------
// Author of this file: christian@cmdev.de

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using System.Drawing.Text;

namespace FontImport
{
    class CMGFontChar
    {
        private CMGFont _parentFont;

        private char _char = ' ';
        public char Char
        { get { return _char; } }

        // must be public because Size is a struct
        public Size Size = new Size( 0, 0 );
        
        private int _lineSizeBytes
        { get { return _parentFont.LineSizeBytes; } }

        private List<List<bool>> _pixelData = new List<List<bool>>();
        public List<List<bool>> PixelData
        { get { return _pixelData; } }


        static public Size MeasureChar( char c, Font windowsFont )
        {
            // measure char (with dummy graphics)
            SizeF size = Graphics.FromImage( new Bitmap( 1, 1 ) ).MeasureString( c.ToString(), windowsFont );
            return new Size( (int)( Math.Ceiling( size.Width ) ), (int)( Math.Ceiling( size.Height ) ) );
        }


        public CMGFontChar( CMGFont parentFont, char c )
        {
            _parentFont = parentFont;
            _char = c;
            
            // get size
            Size = MeasureChar( c, _parentFont.WindowsFont );

            // create image with char
            Bitmap image = new Bitmap( _lineSizeBytes * 8, _parentFont.MaxCharSize.Height );
            Graphics gr = Graphics.FromImage( image );
            
            // set no antialiasing
            gr.SmoothingMode = SmoothingMode.None;
            gr.TextRenderingHint = TextRenderingHint.SingleBitPerPixelGridFit;
            
            // clear background and draw text
            gr.FillRectangle( Brushes.White, new Rectangle( new Point( 0, 0 ), image.Size ) );
            gr.DrawString( c.ToString(), _parentFont.WindowsFont, Brushes.Black, new PointF( 0, 0 ) );

            // for each line
            for ( int y = 0; y < _parentFont.MaxCharSize.Height; y++ )
            {
                // create new line
                List<bool> linePixelData = new List<bool>();
                _pixelData.Add( linePixelData );

                // for each pixel in line
                for ( int x = 0; x < ( _lineSizeBytes * 8 ); x++ )
                    linePixelData.Add( image.GetPixel( x, y ).GetBrightness() < 0.5f );
            }
        }


        public void Optimize()
        {
            // search left char start
            for ( int col = 0; col < ( _lineSizeBytes * 8 ); col++ )
            {
                // break if char starts
                if ( _IsColumnPopulated( 0 ) )
                    break;

                // rotate left char data
                foreach ( List<bool> lineList in _pixelData )
                {
                    lineList.RemoveAt( 0 );
                    lineList.Add( false );
                }
            }

            // search char width (last populated column)
            int rightmostCol = -1;
            for ( int col = 0; col < ( _lineSizeBytes * 8 ); col++ )
                if ( _IsColumnPopulated( col ) )
                    rightmostCol = col;

            // if empty char leave default whitespace
            if ( rightmostCol == -1 )
                return;

            // set width
            Size.Width = rightmostCol + 1;
        }


        private bool _IsColumnPopulated( int col )
        {
            // for each line
            foreach ( List<bool> lineList in _pixelData )
                // check pixel in column
                if ( lineList[ col ] )
                    return true;
            
            return false;
        }


        public List<byte> CompileChar()
        {
            List<byte> rawData = new List<byte>();

            // add size
            rawData.Add( (byte) Size.Width );

            // for each line
            for ( int y = 0; y < _pixelData.Count; y++ )
            {
                // for each byte in line (width/8)
                for ( int colByte = 0; colByte < _lineSizeBytes; colByte++ )
                {
                    byte val = 0;

                    // build byte
                    for ( int bit = 0; bit < 8; bit++ )
                        val = (byte)( ( val << 1 ) | ( _pixelData[ y ][ ( colByte * 8 ) + bit ] ? 1 : 0 ) ); 

                    // add byte to result
                    rawData.Add( val );
                }
            }
            // return result
            return rawData;
        }
    }

}
