//------------------------------------------------------------------------------------------------- 
// Copyright (c) 2oo7, 2oo8 CmDev, Christian Merkle. All rights reserved.       http://www.cmdev.de
//-------------------------------------------------------------------------------------------------
// Author of this file: christian@cmdev.de

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Text;
using System.ComponentModel;


namespace FontImport
{
    /// <summary>
    /// Form to select a font to convert to CMG
    /// </summary>
    public partial class FontSelector : Form
    {
        private object[] backgroundWorkerPreviewStartArgs = null;


        public FontSelector()
        {
            InitializeComponent();
        }


        private void FontSelector_Load( object sender, EventArgs e )
        {
            // fill fonts box
            InstalledFontCollection fonts = new InstalledFontCollection();
            foreach ( FontFamily fontFamily in fonts.Families )
                uxName.Items.Add( fontFamily.Name );

            uxName.SelectedItem = this.Font.Name;

            _ShowPreview();
        }


        private void uxQuit_Click( object sender, EventArgs e )
        {
            // quit application
            Close();
        }


        private Font GetFont()
        {
            return new Font( 
                (string) uxName.SelectedItem, 
                float.Parse( uxSize.Text ),
                ( uxBold.Checked ? FontStyle.Bold : FontStyle.Regular ) | ( uxItalic.Checked ? FontStyle.Italic : FontStyle.Regular )
              );
        }


        /// <summary>
        /// Shows a preview of the selected font
        /// </summary>
        private void _ShowPreview()
        {
            try
            {
                uxPreview.ForeColor = SystemColors.ControlText;
                uxPreview.Font = GetFont();
                uxPreview.Text = CMGFont.GetStringFromRange( int.Parse( uxFrom.Text ), int.Parse( uxTo.Text ) );
                
                // start preview or queue if it is pending
                uxStatistics.Text = "Creating preview, please wait...";
                uxStatisticsOptimized.Text = "Creating preview, please wait...";
                if ( !backgroundWorkerPreview.IsBusy )
                    backgroundWorkerPreview.RunWorkerAsync( new object[] { GetFont(), int.Parse( uxFrom.Text ), int.Parse( uxTo.Text ) } );
                else
                    backgroundWorkerPreviewStartArgs = new object[] { GetFont(), int.Parse( uxFrom.Text ), int.Parse( uxTo.Text ) };

            }
            catch ( Exception ex )
            {
                // an error occured while previewing
                uxPreview.ForeColor = Color.Red;
                uxPreview.Font = this.Font;
                uxPreview.Text = string.Format( "Cannot preview text because an error occured: {0}{1}", Environment.NewLine, ex.ToString() );
            }
        }

        private void _ShowPreview( object sender, EventArgs e )
        {
            _ShowPreview();
        }
        
        private void backgroundWorkerPreview_DoWork( object sender, DoWorkEventArgs e )
        {   
            object[] args = e.Argument as object[];
            if ( args != null )
                e.Result = new CMGFont( (Font) args[0], (int) args[1], (int) args[2], false );
        }

        private void backgroundWorkerPreview_RunWorkerCompleted( object sender, RunWorkerCompletedEventArgs e )
        {
            CMGFont font = e.Result as CMGFont;
            if ( font != null )
            {
                uxStatistics.Text = font.ToString();
                font.Optimize();
                uxStatisticsOptimized.Text = font.ToString();
            }

            // start queued preview
            if ( backgroundWorkerPreviewStartArgs != null )
            {
                uxStatistics.Text = "Creating preview, please wait...";
                uxStatisticsOptimized.Text = "Creating preview, please wait...";
                backgroundWorkerPreview.RunWorkerAsync( backgroundWorkerPreviewStartArgs );
                backgroundWorkerPreviewStartArgs = null;
            }
        }



        private void uxExport_Click( object sender, EventArgs e )
        {
            try
            {
                CMGFont font = new CMGFont( GetFont(), int.Parse( uxFrom.Text ), int.Parse( uxTo.Text ), true );
                font.Optimize();
                string cCode = font.CompileToCCode();
                CodeResult codeResult = new CodeResult( cCode, "CMG_Font_" + font.FontName.Replace( ' ', '_' ) );
                codeResult.Show();
            }
            catch ( Exception ex )
            {
                MessageBox.Show( string.Format( "Error while converting the font:{0}{1}", Environment.NewLine, ex.ToString() ), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error );
            }
        }

        private void DigitTextBox_KeyPress( object sender, KeyPressEventArgs e )
        {
            // only digits allowed
            e.Handled = !( char.IsDigit( e.KeyChar ) || char.IsControl( e.KeyChar ) );
        }




    }
}