//***************************************************************************
// ARMStartup.c
//***************************************************************************
// CMG - Christian Merkle Graphics
// Copyright (C) 2oo6, 2oo7 Christian Merkle 
// All rights reserved.
//
// To contact Christian Merkle, see his website at:
// http://www.cmerkle.de
//***************************************************************************
// Author:          Christian Merkle, Hubert Hoegl
// Created:         09.02.2007
//***************************************************************************
// History:
//   09.02.2006     *stable*
//***************************************************************************
// ToDo:
//   [ ]    xxx
//***************************************************************************

#include "lpc22xx.h"


// exports:
void ARMStartup( void );




/* System Config */
/* from Hubert Hoegl */

/* Olimex LPC-P2106
	PLL
	
   - Main clock F_OSC=14,7MHz @ Olimex LPC-P2106) [limits: 10 MHz to 25 MHz]
   - System should run at max. Frequency (60MHz) [limit: max 60 MHz]
   - Choose multiplier M=4 
     so cclk = M * F_OSC= 4 * 14745000Hz = 58980000 Hz
   - MSEL-Bit in PLLCFG (bits 0-4) MSEL = M-1
   - F_CCO must be inbetween the limits 156 MHz to 320 MHz
     datasheet: F_CCO = F_OSC * M * 2 * P
   - choose devider P=2 => F_CCO = 14745000Hz * 4 * 2 * 2
     = 235920000 ~=236 MHz
   - PSEL0 (Bit5 in PLLCFG) = 0, PSEL1 (Bit6) = 1 
*/

/* Embest board:
 * F_OSC = 10.0 MHz
 * Processor clock CCLK = PLL_M * F_OSC = 60 MHz
 * MSEL = 6 - 1 = 5
 * PSEL[1:0] = 0b10 (0x2)
 * F_CCO = F_OSC * PLL_M * 2 * PSEL = 10 MHz * 6 * 2 * 2 = 240 MHz
 * 
 */

/* #define FOSC		14745000 */
#define FOSC		10000000 /* Embest has a 10 MHz crystal */
#define PLL_M		6	
#define MSEL		(PLL_M-1)  /* bit positions [4:0] in PLLCFG */
#define PSEL0 		5  /* bit position in PLLCFG */
#define PSEL1 		6  /* bit position in PLLCFG */

#define PLLE		0  /* bit pos of PLLE bit in PLLCON */ 
#define PLLC		1  /* bit pos of PLLC bit in PLLCON */ 

#define PLOCK		10 /* bit position of PLOCK in PLLSTAT */

#define PLL_FEED1	0xAA
#define PLL_FEED2	0x55


/*
	MAM(Memory Accelerator Module)
	- choosen: MAM fully enabled = MAM-Mode 2
	- System-Clock cclk=59MHz -> 3 CCLKs are proposed as fetch timing
*/
#define MAM_MODE 	2
#define MAM_FETCH   3
	
/*
	VPB (V... Pheriphal Bus)
	- choosen: VPB should run at full speed -> devider VPBDIV=1
	=> pclk = cclk = 59MHz
*/
#define VPBDIV_VAL	1

/* 
	SCB
*/
#define MEMMAP_BOOT_LOADER_MODE   0       // Interrupt vectors are re-mapped to Boot Block.
#define MEMMAP_USER_FLASH_MODE    (1<<0)  // Interrupt vectors are not re-mapped and reside in Flash.
#define MEMMAP_USER_RAM_MODE      (1<<1)  // Interrupt vectors are re-mapped to Static RAM.




void ARMStartup( void )
{
	// --- enable and connect the PLL (Phase Locked Loop) ---
	// a. set multiplier and divider
	PLLCFG = MSEL | (1<<PSEL1) | (0<<PSEL0);
	// b. enable PLL
	PLLCON = (1<<PLLE);
	// c. feed sequence
	PLLFEED = PLL_FEED1;
	PLLFEED = PLL_FEED2;
	// d. wait for PLL lock (PLOCK bit is set if locked)
	while (!(PLLSTAT & (1<<PLOCK)));
	// e. connect (and enable) PLL
	PLLCON = (1<<PLLE) | (1<<PLLC);
	// f. feed sequence
	PLLFEED = PLL_FEED1;
	PLLFEED = PLL_FEED2;
	
	// --- setup and enable the MAM (Memory Accelerator Module) ---
	// a. start change by turning of the MAM (redundant)
	MAMCR = 0;	
	// b. set MAM-Fetch cycle to 3 cclk as recommended for >40MHz
	MAMTIM = MAM_FETCH;
	// c. enable MAM 
	MAMCR = MAM_MODE;
	
	// --- set VPB speed ---
	VPBDIV = VPBDIV_VAL;
	
	// --- map INT-vector ---
	#if defined(RAM_RUN)
        MEMMAP = MEMMAP_USER_RAM_MODE;
	#elif defined(ROM_RUN)
        MEMMAP = MEMMAP_USER_FLASH_MODE;
	#else
        #error RUN_MODE not defined!
	#endif
}
